/* -*- mode: c; coding: utf-8 -*- */

#include <IceC/IceC.h>
#include <IceC/IceUtil.h>

typedef struct IceStorm_QoS_Pair {
    Ice_String key;
    Ice_String value;
} IceStorm_QoS_Pair;

typedef IceStorm_QoS_Pair* IceStorm_QoS_PairPtr;

typedef struct IceStorm_QoS {
    byte size;
    IceStorm_QoS_PairPtr items;
} IceStorm_QoS;

typedef IceStorm_QoS* IceStorm_QoSPtr;

#define IceStorm_QoS_init Ice_Dict_init

#define IceStorm_QoS_set(DICT, INDEX, KEY, VALUE)       \
do {                                                    \
    Ice_String_init(DICT.items[INDEX].key, KEY);        \
    Ice_String_init(DICT.items[INDEX].value, VALUE);    \
} while(0)                                              

static void
IceStorm_QoS_writeToOutputStream(IceStorm_QoSPtr self,
                                 Ice_OutputStreamPtr os) {
    byte i = 0;

    trace();
    
    Ptr_check(os);
    
    Ice_OutputStream_writeSize(os, self->size);
    
    for (i=0; i<self->size; i++) {
        Ice_OutputStream_writeIceString(os, self->items[i].key);
        Ice_OutputStream_writeIceString(os, self->items[i].value);
    }
}

static void
IceStorm_QoS_readFromInputStream(IceStorm_QoSPtr self,
                                 Ice_InputStreamPtr is) {
    Ice_Int size = 0;
    byte i = 0;

    trace();
    
    Ptr_check(is);
    
    Ice_InputStream_readSize(is, &size, true);
    assert(size < 256);
    self->size = size;
    
    for (i=0; i<self->size; i++) {
        Ice_InputStream_readIceString(is, &(self->items[i].key));
        Ice_InputStream_readIceString(is, &(self->items[i].value));
    }
}

typedef struct IceStorm_Topic {
    Ice_Object __base;
} IceStorm_Topic;

typedef IceStorm_Topic* IceStorm_TopicPtr;

typedef Ice_ObjectPrx IceStorm_TopicPrx;
typedef Ice_ObjectPrx* IceStorm_TopicPrxPtr;

/* forward declaration */
static void
IceStorm_Topic_methodHandler(IceStorm_TopicPtr self,
                             Ice_InputStreamPtr is,
                             Ice_OutputStreamPtr os);

static void
IceStorm_Topic_init(IceStorm_TopicPtr self) {

    trace();
    
    Ice_Object_init((Ice_ObjectPtr)self,
            (Ice_MethodHandlerPtr)IceStorm_Topic_methodHandler);
}

Ice_String
IceStorm_TopicI_getName(IceStorm_TopicPtr self) __attribute__((weak));

void
IceStorm_TopicI_getPublisher(IceStorm_TopicPtr self,
                             Ice_ObjectPrxPtr retval) __attribute__((weak));

void
IceStorm_TopicI_subscribeAndGetPublisher(IceStorm_TopicPtr self,
                                         IceStorm_QoS theQoS,
                                         Ice_ObjectPrxPtr subscriber,
                                         Ice_ObjectPrxPtr retval) __attribute__((weak));

void
IceStorm_TopicI_unsubscribe(IceStorm_TopicPtr self,
                            Ice_ObjectPrxPtr subscriber) __attribute__((weak));

void
IceStorm_TopicI_destroy(IceStorm_TopicPtr self) __attribute__((weak));

static void
IceStorm_Topic_getName__dispatch(IceStorm_TopicPtr self,
                                 Ice_InputStreamPtr is,
                                 Ice_OutputStreamPtr os) {
    Ice_String retval;

    trace();
    
    if (&IceStorm_TopicI_getName == 0) {
        return;
    }
    retval = IceStorm_TopicI_getName(self);
    Ice_OutputStream_writeIceString(os, retval);
}

static void
IceStorm_Topic_getPublisher__dispatch(IceStorm_TopicPtr self,
                                      Ice_InputStreamPtr is,
                                      Ice_OutputStreamPtr os) {
    Ice_ObjectPrx retval;

    trace();
    
    if (&IceStorm_TopicI_getPublisher == 0) {
        return;
    }
    IceStorm_TopicI_getPublisher(self, &retval);
    Ice_ObjectPrx_writeToOutputStream(&(retval), os);
}

static void
IceStorm_Topic_subscribeAndGetPublisher__dispatch(IceStorm_TopicPtr self,
                                                  Ice_InputStreamPtr is,
                                                  Ice_OutputStreamPtr os) {
    IceStorm_QoS theQoS;
    Ice_ObjectPrx subscriber;
    Ice_ObjectPrx retval;

    trace();
    
    if (&IceStorm_TopicI_subscribeAndGetPublisher == 0) {
        return;
    }
    
    /* allocate space for dictionary */
    Ice_InputStream_readSize(is, (Ice_Int*)&theQoS.size, false);
    byte theQoS_items_data[sizeof(IceStorm_QoS_Pair) * theQoS.size];
    theQoS.items = (IceStorm_QoS_Pair*)theQoS_items_data;
    IceStorm_QoS_readFromInputStream(&theQoS, is);
    
    Ice_ObjectPrx_readFromInputStream(&subscriber, is);
    
    IceStorm_TopicI_subscribeAndGetPublisher(self, theQoS, &subscriber, &retval);
    Ice_ObjectPrx_writeToOutputStream(&(retval), os);
}

static void
IceStorm_Topic_unsubscribe__dispatch(IceStorm_TopicPtr self,
                                     Ice_InputStreamPtr is,
                                     Ice_OutputStreamPtr os) {
    Ice_ObjectPrx subscriber;

    trace();
    
    if (&IceStorm_TopicI_unsubscribe == 0) {
        return;
    }
    Ice_ObjectPrx_readFromInputStream(&subscriber, is);
    
    IceStorm_TopicI_unsubscribe(self, &subscriber);
}

static void
IceStorm_Topic_destroy__dispatch(IceStorm_TopicPtr self,
                                 Ice_InputStreamPtr is,
                                 Ice_OutputStreamPtr os) {

    trace();
    
    if (&IceStorm_TopicI_destroy == 0) {
        return;
    }
    IceStorm_TopicI_destroy(self);
}

static void
IceStorm_Topic_methodHandler(IceStorm_TopicPtr self,
                             Ice_InputStreamPtr is,
                             Ice_OutputStreamPtr os) {
    Ice_Int operationSize = 0;
    char operation[MAX_OPERATION_NAME_SIZE];
    byte mode = 0;
    byte contextSize = 0;
    Ice_Int encapSize = 0;
    byte major = 0;
    byte minor = 0;

    trace();
    
    Ptr_check(self);
    Ptr_check(is);
    Ptr_check(os);
    
    Ice_InputStream_readSize(is, &operationSize, false);
    assert(operationSize <= MAX_OPERATION_NAME_SIZE);
    
    Ice_InputStream_readString(is, operation);
    
    /* FIXME: operationMode ignored */
    Ice_InputStream_readByte(is, &mode);
    
    /* FIXME: Context not supported */
    Ice_InputStream_readByte(is, &contextSize);
    assert(contextSize == 0);
    
    Ice_InputStream_readInt(is, &encapSize);
    Ice_InputStream_readByte(is, &major);
    Ice_InputStream_readByte(is, &minor);
    assert(major == 1 && minor == 0);
    assert(encapSize - 6 == (is->size - (is->next - is->data)));
    
    if (strcmp(operation, "getName") == 0) {
        IceStorm_Topic_getName__dispatch(self, is, os);
        return;
    }
    
    if (strcmp(operation, "getPublisher") == 0) {
        IceStorm_Topic_getPublisher__dispatch(self, is, os);
        return;
    }
    
    if (strcmp(operation, "subscribeAndGetPublisher") == 0) {
        IceStorm_Topic_subscribeAndGetPublisher__dispatch(self, is, os);
        return;
    }
    
    if (strcmp(operation, "unsubscribe") == 0) {
        IceStorm_Topic_unsubscribe__dispatch(self, is, os);
        return;
    }
    
    if (strcmp(operation, "destroy") == 0) {
        IceStorm_Topic_destroy__dispatch(self, is, os);
        return;
    }
    
    c_throw("Ice_OperationNotExistException", __FILE__, __LINE__);
}

static Ice_String
IceStorm_Topic_getName(IceStorm_TopicPrxPtr self) {
    Ice_OutputStreamPtr os = &(self->stream);
    Ice_String retval;

    trace();
    
    Ptr_check(self);
    
    Ice_ObjectPrx_connect(self);
    Ice_OutputStream_init(os);
    
    /* request header */
    Ice_OutputStream_writeBlob(os, Ice_header, sizeof(Ice_header));
    
    /* request body */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 1);
    }
    if (self->mode != Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 0);
    }
    
    Ice_OutputStream_writeIdentity(os, self->identity);
    Ice_OutputStream_writeString(os, "");
    Ice_OutputStream_writeString(os, "getName");
    Ice_OutputStream_writeByte(os, Ice_Idempotent);
    Ice_OutputStream_writeByte(os, 0);
    
    /* encapsulated params */
    Ice_OutputStream_startWriteEncaps(os);
    Ice_OutputStream_endWriteEncaps(os);
    Ice_OutputStream_setMessageSize(os);
    
    Ice_ObjectPrx_send(self);
    
    /* parse reply message (only on twoway invocations) */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_InputStream is;
        char expected[10] = {'I', 'c', 'e', 'P', 1, 0, 1, 0, Ice_replyMsg, 0};
        Ice_Int size = 0;
        Ice_Int requestId = 0;
        Ice_Byte replyStatus = 0;
        Ice_Int encapsSize = 0;
        Ice_Byte skipByte = 0;
        
        uint16_t timeout = 5000;
        bool ready = false;
        while (!ready) {
            IcePlugin_Connection_dataReady(&(self->connection), &ready);
            IceUtil_delay(1);
            if (! --timeout) {
                c_throw("Ice_TimeoutException", __FILE__, __LINE__);
                return retval;
            }
        }
        
        /* reply header */
        Ice_InputStream_init(&is, self->connection.fd, self->connection.epinfo->type);
        if (memcmp(expected, is.next, 10) != 0) {
            c_throw("Ice_ProtocolException", __FILE__, __LINE__);
            return retval;
        }
        
        /* skip header from input stream */
        is.next += 10;
        Ice_InputStream_readInt(&is, &size);
            if (size > MAX_MESSAGE_SIZE) {
            c_throw("Ice_MemoryLimitException", __FILE__, __LINE__);
            return retval;
        }
        
        /* reply body */
        Ice_InputStream_readInt(&is, &requestId);
            Ice_InputStream_readByte(&is, &replyStatus);
            
        /* check replyStatus */
        switch (replyStatus) {
        case 0: {
            /* success */
            break;
        }
        case 1: {
            /* user exception */
            Ice_ObjectPrx_parseUserException(self, &is);
            return retval;
        }
        case 2: {
            c_throw("Ice_ObjectDoesNotExist", __FILE__, __LINE__);
            return retval;
        }
        case 3: {
            c_throw("Ice_FacetDoesNotExist", __FILE__, __LINE__);
            return retval;
        }
        case 4: {
            c_throw("Ice_OperationDoesNotExist", __FILE__, __LINE__);
            return retval;
        }
        case 5:
        case 6:
        case 7: {
            /* unkown exception (reason as string) */
            /* FIXME: unmarshall and print reason */
            c_throw("Ice_UnknownException", __FILE__, __LINE__);
            return retval;
        }}
        
        Ice_InputStream_readInt(&is, &encapsSize);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readIceString(&is, &retval);
        }
    
    Ice_ObjectPrx_closeConnection(self);
    return retval;
}

static void
IceStorm_Topic_getPublisher(IceStorm_TopicPrxPtr self,
                            Ice_ObjectPrxPtr retval) {
    Ice_OutputStreamPtr os = &(self->stream);

    trace();
    
    Ptr_check(self);
    
    Ice_ObjectPrx_connect(self);
    Ice_OutputStream_init(os);
    
    /* request header */
    Ice_OutputStream_writeBlob(os, Ice_header, sizeof(Ice_header));
    
    /* request body */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 1);
    }
    if (self->mode != Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 0);
    }
    
    Ice_OutputStream_writeIdentity(os, self->identity);
    Ice_OutputStream_writeString(os, "");
    Ice_OutputStream_writeString(os, "getPublisher");
    Ice_OutputStream_writeByte(os, Ice_Idempotent);
    Ice_OutputStream_writeByte(os, 0);
    
    /* encapsulated params */
    Ice_OutputStream_startWriteEncaps(os);
    Ice_OutputStream_endWriteEncaps(os);
    Ice_OutputStream_setMessageSize(os);
    
    Ice_ObjectPrx_send(self);
    
    /* parse reply message (only on twoway invocations) */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_InputStream is;
        char expected[10] = {'I', 'c', 'e', 'P', 1, 0, 1, 0, Ice_replyMsg, 0};
        Ice_Int size = 0;
        Ice_Int requestId = 0;
        Ice_Byte replyStatus = 0;
        Ice_Int encapsSize = 0;
        Ice_Byte skipByte = 0;
        
        uint16_t timeout = 5000;
        bool ready = false;
        while (!ready) {
            IcePlugin_Connection_dataReady(&(self->connection), &ready);
            IceUtil_delay(1);
            if (! --timeout) {
                c_throw("Ice_TimeoutException", __FILE__, __LINE__);
                return;
            }
        }
        
        /* reply header */
        Ice_InputStream_init(&is, self->connection.fd, self->connection.epinfo->type);
        if (memcmp(expected, is.next, 10) != 0) {
            c_throw("Ice_ProtocolException", __FILE__, __LINE__);
            return;
        }
        
        /* skip header from input stream */
        is.next += 10;
        Ice_InputStream_readInt(&is, &size);
            if (size > MAX_MESSAGE_SIZE) {
            c_throw("Ice_MemoryLimitException", __FILE__, __LINE__);
            return;
        }
        
        /* reply body */
        Ice_InputStream_readInt(&is, &requestId);
            Ice_InputStream_readByte(&is, &replyStatus);
            
        /* check replyStatus */
        switch (replyStatus) {
        case 0: {
            /* success */
            break;
        }
        case 1: {
            /* user exception */
            Ice_ObjectPrx_parseUserException(self, &is);
            return;
        }
        case 2: {
            c_throw("Ice_ObjectDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 3: {
            c_throw("Ice_FacetDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 4: {
            c_throw("Ice_OperationDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 5:
        case 6:
        case 7: {
            /* unkown exception (reason as string) */
            /* FIXME: unmarshall and print reason */
            c_throw("Ice_UnknownException", __FILE__, __LINE__);
            return;
        }}
        
        Ice_InputStream_readInt(&is, &encapsSize);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_ObjectPrx_readFromInputStream(retval, &is);
        }
    
    Ice_ObjectPrx_closeConnection(self);
}

static void
IceStorm_Topic_subscribeAndGetPublisher(IceStorm_TopicPrxPtr self,
                                        IceStorm_QoS theQoS,
                                        Ice_ObjectPrxPtr subscriber,
                                        Ice_ObjectPrxPtr retval) {
    Ice_OutputStreamPtr os = &(self->stream);

    trace();
    
    Ptr_check(self);
    
    Ice_ObjectPrx_connect(self);
    Ice_OutputStream_init(os);
    
    /* request header */
    Ice_OutputStream_writeBlob(os, Ice_header, sizeof(Ice_header));
    
    /* request body */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 1);
    }
    if (self->mode != Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 0);
    }
    
    Ice_OutputStream_writeIdentity(os, self->identity);
    Ice_OutputStream_writeString(os, "");
    Ice_OutputStream_writeString(os, "subscribeAndGetPublisher");
    Ice_OutputStream_writeByte(os, Ice_Normal);
    Ice_OutputStream_writeByte(os, 0);
    
    /* encapsulated params */
    Ice_OutputStream_startWriteEncaps(os);
    IceStorm_QoS_writeToOutputStream(&(theQoS), os);
    Ice_ObjectPrx_writeToOutputStream(subscriber, os);
    Ice_OutputStream_endWriteEncaps(os);
    Ice_OutputStream_setMessageSize(os);
    
    Ice_ObjectPrx_send(self);
    
    /* parse reply message (only on twoway invocations) */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_InputStream is;
        char expected[10] = {'I', 'c', 'e', 'P', 1, 0, 1, 0, Ice_replyMsg, 0};
        Ice_Int size = 0;
        Ice_Int requestId = 0;
        Ice_Byte replyStatus = 0;
        Ice_Int encapsSize = 0;
        Ice_Byte skipByte = 0;
        
        uint16_t timeout = 5000;
        bool ready = false;
        while (!ready) {
            IcePlugin_Connection_dataReady(&(self->connection), &ready);
            IceUtil_delay(1);
            if (! --timeout) {
                c_throw("Ice_TimeoutException", __FILE__, __LINE__);
                return;
            }
        }
        
        /* reply header */
        Ice_InputStream_init(&is, self->connection.fd, self->connection.epinfo->type);
        if (memcmp(expected, is.next, 10) != 0) {
            c_throw("Ice_ProtocolException", __FILE__, __LINE__);
            return;
        }
        
        /* skip header from input stream */
        is.next += 10;
        Ice_InputStream_readInt(&is, &size);
            if (size > MAX_MESSAGE_SIZE) {
            c_throw("Ice_MemoryLimitException", __FILE__, __LINE__);
            return;
        }
        
        /* reply body */
        Ice_InputStream_readInt(&is, &requestId);
            Ice_InputStream_readByte(&is, &replyStatus);
            
        /* check replyStatus */
        switch (replyStatus) {
        case 0: {
            /* success */
            break;
        }
        case 1: {
            /* user exception */
            Ice_ObjectPrx_parseUserException(self, &is);
            return;
        }
        case 2: {
            c_throw("Ice_ObjectDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 3: {
            c_throw("Ice_FacetDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 4: {
            c_throw("Ice_OperationDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 5:
        case 6:
        case 7: {
            /* unkown exception (reason as string) */
            /* FIXME: unmarshall and print reason */
            c_throw("Ice_UnknownException", __FILE__, __LINE__);
            return;
        }}
        
        Ice_InputStream_readInt(&is, &encapsSize);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_ObjectPrx_readFromInputStream(retval, &is);
        }
    
    Ice_ObjectPrx_closeConnection(self);
}

static void
IceStorm_Topic_unsubscribe(IceStorm_TopicPrxPtr self,
                           Ice_ObjectPrxPtr subscriber) {
    Ice_OutputStreamPtr os = &(self->stream);

    trace();
    
    Ptr_check(self);
    
    Ice_ObjectPrx_connect(self);
    Ice_OutputStream_init(os);
    
    /* request header */
    Ice_OutputStream_writeBlob(os, Ice_header, sizeof(Ice_header));
    
    /* request body */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 1);
    }
    if (self->mode != Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 0);
    }
    
    Ice_OutputStream_writeIdentity(os, self->identity);
    Ice_OutputStream_writeString(os, "");
    Ice_OutputStream_writeString(os, "unsubscribe");
    Ice_OutputStream_writeByte(os, Ice_Idempotent);
    Ice_OutputStream_writeByte(os, 0);
    
    /* encapsulated params */
    Ice_OutputStream_startWriteEncaps(os);
    Ice_ObjectPrx_writeToOutputStream(subscriber, os);
    Ice_OutputStream_endWriteEncaps(os);
    Ice_OutputStream_setMessageSize(os);
    
    Ice_ObjectPrx_send(self);
    
    /* parse reply message (only on twoway invocations) */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_InputStream is;
        char expected[10] = {'I', 'c', 'e', 'P', 1, 0, 1, 0, Ice_replyMsg, 0};
        Ice_Int size = 0;
        Ice_Int requestId = 0;
        Ice_Byte replyStatus = 0;
        Ice_Int encapsSize = 0;
        Ice_Byte skipByte = 0;
        
        uint16_t timeout = 5000;
        bool ready = false;
        while (!ready) {
            IcePlugin_Connection_dataReady(&(self->connection), &ready);
            IceUtil_delay(1);
            if (! --timeout) {
                c_throw("Ice_TimeoutException", __FILE__, __LINE__);
                return;
            }
        }
        
        /* reply header */
        Ice_InputStream_init(&is, self->connection.fd, self->connection.epinfo->type);
        if (memcmp(expected, is.next, 10) != 0) {
            c_throw("Ice_ProtocolException", __FILE__, __LINE__);
            return;
        }
        
        /* skip header from input stream */
        is.next += 10;
        Ice_InputStream_readInt(&is, &size);
            if (size > MAX_MESSAGE_SIZE) {
            c_throw("Ice_MemoryLimitException", __FILE__, __LINE__);
            return;
        }
        
        /* reply body */
        Ice_InputStream_readInt(&is, &requestId);
            Ice_InputStream_readByte(&is, &replyStatus);
            
        /* check replyStatus */
        switch (replyStatus) {
        case 0: {
            /* success */
            break;
        }
        case 1: {
            /* user exception */
            Ice_ObjectPrx_parseUserException(self, &is);
            return;
        }
        case 2: {
            c_throw("Ice_ObjectDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 3: {
            c_throw("Ice_FacetDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 4: {
            c_throw("Ice_OperationDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 5:
        case 6:
        case 7: {
            /* unkown exception (reason as string) */
            /* FIXME: unmarshall and print reason */
            c_throw("Ice_UnknownException", __FILE__, __LINE__);
            return;
        }}
        
        Ice_InputStream_readInt(&is, &encapsSize);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readByte(&is, &skipByte);
        }
    
    Ice_ObjectPrx_closeConnection(self);
}

static void
IceStorm_Topic_destroy(IceStorm_TopicPrxPtr self) {
    Ice_OutputStreamPtr os = &(self->stream);

    trace();
    
    Ptr_check(self);
    
    Ice_ObjectPrx_connect(self);
    Ice_OutputStream_init(os);
    
    /* request header */
    Ice_OutputStream_writeBlob(os, Ice_header, sizeof(Ice_header));
    
    /* request body */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 1);
    }
    if (self->mode != Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 0);
    }
    
    Ice_OutputStream_writeIdentity(os, self->identity);
    Ice_OutputStream_writeString(os, "");
    Ice_OutputStream_writeString(os, "destroy");
    Ice_OutputStream_writeByte(os, Ice_Normal);
    Ice_OutputStream_writeByte(os, 0);
    
    /* encapsulated params */
    Ice_OutputStream_startWriteEncaps(os);
    Ice_OutputStream_endWriteEncaps(os);
    Ice_OutputStream_setMessageSize(os);
    
    Ice_ObjectPrx_send(self);
    
    /* parse reply message (only on twoway invocations) */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_InputStream is;
        char expected[10] = {'I', 'c', 'e', 'P', 1, 0, 1, 0, Ice_replyMsg, 0};
        Ice_Int size = 0;
        Ice_Int requestId = 0;
        Ice_Byte replyStatus = 0;
        Ice_Int encapsSize = 0;
        Ice_Byte skipByte = 0;
        
        uint16_t timeout = 5000;
        bool ready = false;
        while (!ready) {
            IcePlugin_Connection_dataReady(&(self->connection), &ready);
            IceUtil_delay(1);
            if (! --timeout) {
                c_throw("Ice_TimeoutException", __FILE__, __LINE__);
                return;
            }
        }
        
        /* reply header */
        Ice_InputStream_init(&is, self->connection.fd, self->connection.epinfo->type);
        if (memcmp(expected, is.next, 10) != 0) {
            c_throw("Ice_ProtocolException", __FILE__, __LINE__);
            return;
        }
        
        /* skip header from input stream */
        is.next += 10;
        Ice_InputStream_readInt(&is, &size);
            if (size > MAX_MESSAGE_SIZE) {
            c_throw("Ice_MemoryLimitException", __FILE__, __LINE__);
            return;
        }
        
        /* reply body */
        Ice_InputStream_readInt(&is, &requestId);
            Ice_InputStream_readByte(&is, &replyStatus);
            
        /* check replyStatus */
        switch (replyStatus) {
        case 0: {
            /* success */
            break;
        }
        case 1: {
            /* user exception */
            Ice_ObjectPrx_parseUserException(self, &is);
            return;
        }
        case 2: {
            c_throw("Ice_ObjectDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 3: {
            c_throw("Ice_FacetDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 4: {
            c_throw("Ice_OperationDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 5:
        case 6:
        case 7: {
            /* unkown exception (reason as string) */
            /* FIXME: unmarshall and print reason */
            c_throw("Ice_UnknownException", __FILE__, __LINE__);
            return;
        }}
        
        Ice_InputStream_readInt(&is, &encapsSize);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readByte(&is, &skipByte);
        }
    
    Ice_ObjectPrx_closeConnection(self);
}

typedef struct IceStorm_TopicManager {
    Ice_Object __base;
} IceStorm_TopicManager;

typedef IceStorm_TopicManager* IceStorm_TopicManagerPtr;

typedef Ice_ObjectPrx IceStorm_TopicManagerPrx;
typedef Ice_ObjectPrx* IceStorm_TopicManagerPrxPtr;

/* forward declaration */
static void
IceStorm_TopicManager_methodHandler(IceStorm_TopicManagerPtr self,
                                    Ice_InputStreamPtr is,
                                    Ice_OutputStreamPtr os);

static void
IceStorm_TopicManager_init(IceStorm_TopicManagerPtr self) {

    trace();
    
    Ice_Object_init((Ice_ObjectPtr)self,
            (Ice_MethodHandlerPtr)IceStorm_TopicManager_methodHandler);
}

void
IceStorm_TopicManagerI_create(IceStorm_TopicManagerPtr self,
                              Ice_String name,
                              IceStorm_TopicPrxPtr retval) __attribute__((weak));

void
IceStorm_TopicManagerI_retrieve(IceStorm_TopicManagerPtr self,
                                Ice_String name,
                                IceStorm_TopicPrxPtr retval) __attribute__((weak));

static void
IceStorm_TopicManager_create__dispatch(IceStorm_TopicManagerPtr self,
                                       Ice_InputStreamPtr is,
                                       Ice_OutputStreamPtr os) {
    Ice_String name;
    IceStorm_TopicPrx retval;

    trace();
    
    if (&IceStorm_TopicManagerI_create == 0) {
        return;
    }
    Ice_InputStream_readIceString(is, &name);
    
    IceStorm_TopicManagerI_create(self, name, &retval);
    Ice_ObjectPrx_writeToOutputStream(&(retval), os);
}

static void
IceStorm_TopicManager_retrieve__dispatch(IceStorm_TopicManagerPtr self,
                                         Ice_InputStreamPtr is,
                                         Ice_OutputStreamPtr os) {
    Ice_String name;
    IceStorm_TopicPrx retval;

    trace();
    
    if (&IceStorm_TopicManagerI_retrieve == 0) {
        return;
    }
    Ice_InputStream_readIceString(is, &name);
    
    IceStorm_TopicManagerI_retrieve(self, name, &retval);
    Ice_ObjectPrx_writeToOutputStream(&(retval), os);
}

static void
IceStorm_TopicManager_methodHandler(IceStorm_TopicManagerPtr self,
                                    Ice_InputStreamPtr is,
                                    Ice_OutputStreamPtr os) {
    Ice_Int operationSize = 0;
    char operation[MAX_OPERATION_NAME_SIZE];
    byte mode = 0;
    byte contextSize = 0;
    Ice_Int encapSize = 0;
    byte major = 0;
    byte minor = 0;

    trace();
    
    Ptr_check(self);
    Ptr_check(is);
    Ptr_check(os);
    
    Ice_InputStream_readSize(is, &operationSize, false);
    assert(operationSize <= MAX_OPERATION_NAME_SIZE);
    
    Ice_InputStream_readString(is, operation);
    
    /* FIXME: operationMode ignored */
    Ice_InputStream_readByte(is, &mode);
    
    /* FIXME: Context not supported */
    Ice_InputStream_readByte(is, &contextSize);
    assert(contextSize == 0);
    
    Ice_InputStream_readInt(is, &encapSize);
    Ice_InputStream_readByte(is, &major);
    Ice_InputStream_readByte(is, &minor);
    assert(major == 1 && minor == 0);
    assert(encapSize - 6 == (is->size - (is->next - is->data)));
    
    if (strcmp(operation, "create") == 0) {
        IceStorm_TopicManager_create__dispatch(self, is, os);
        return;
    }
    
    if (strcmp(operation, "retrieve") == 0) {
        IceStorm_TopicManager_retrieve__dispatch(self, is, os);
        return;
    }
    
    c_throw("Ice_OperationNotExistException", __FILE__, __LINE__);
}

static void
IceStorm_TopicManager_create(IceStorm_TopicManagerPrxPtr self,
                             Ice_String name,
                             IceStorm_TopicPrxPtr retval) {
    Ice_OutputStreamPtr os = &(self->stream);

    trace();
    
    Ptr_check(self);
    
    Ice_ObjectPrx_connect(self);
    Ice_OutputStream_init(os);
    
    /* request header */
    Ice_OutputStream_writeBlob(os, Ice_header, sizeof(Ice_header));
    
    /* request body */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 1);
    }
    if (self->mode != Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 0);
    }
    
    Ice_OutputStream_writeIdentity(os, self->identity);
    Ice_OutputStream_writeString(os, "");
    Ice_OutputStream_writeString(os, "create");
    Ice_OutputStream_writeByte(os, Ice_Normal);
    Ice_OutputStream_writeByte(os, 0);
    
    /* encapsulated params */
    Ice_OutputStream_startWriteEncaps(os);
    Ice_OutputStream_writeIceString(os, name);
    Ice_OutputStream_endWriteEncaps(os);
    Ice_OutputStream_setMessageSize(os);
    
    Ice_ObjectPrx_send(self);
    
    /* parse reply message (only on twoway invocations) */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_InputStream is;
        char expected[10] = {'I', 'c', 'e', 'P', 1, 0, 1, 0, Ice_replyMsg, 0};
        Ice_Int size = 0;
        Ice_Int requestId = 0;
        Ice_Byte replyStatus = 0;
        Ice_Int encapsSize = 0;
        Ice_Byte skipByte = 0;
        
        uint16_t timeout = 5000;
        bool ready = false;
        while (!ready) {
            IcePlugin_Connection_dataReady(&(self->connection), &ready);
            IceUtil_delay(1);
            if (! --timeout) {
                c_throw("Ice_TimeoutException", __FILE__, __LINE__);
                return;
            }
        }
        
        /* reply header */
        Ice_InputStream_init(&is, self->connection.fd, self->connection.epinfo->type);
        if (memcmp(expected, is.next, 10) != 0) {
            c_throw("Ice_ProtocolException", __FILE__, __LINE__);
            return;
        }
        
        /* skip header from input stream */
        is.next += 10;
        Ice_InputStream_readInt(&is, &size);
            if (size > MAX_MESSAGE_SIZE) {
            c_throw("Ice_MemoryLimitException", __FILE__, __LINE__);
            return;
        }
        
        /* reply body */
        Ice_InputStream_readInt(&is, &requestId);
            Ice_InputStream_readByte(&is, &replyStatus);
            
        /* check replyStatus */
        switch (replyStatus) {
        case 0: {
            /* success */
            break;
        }
        case 1: {
            /* user exception */
            Ice_ObjectPrx_parseUserException(self, &is);
            return;
        }
        case 2: {
            c_throw("Ice_ObjectDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 3: {
            c_throw("Ice_FacetDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 4: {
            c_throw("Ice_OperationDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 5:
        case 6:
        case 7: {
            /* unkown exception (reason as string) */
            /* FIXME: unmarshall and print reason */
            c_throw("Ice_UnknownException", __FILE__, __LINE__);
            return;
        }}
        
        Ice_InputStream_readInt(&is, &encapsSize);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_ObjectPrx_readFromInputStream(retval, &is);
        }
    
    Ice_ObjectPrx_closeConnection(self);
}

static void
IceStorm_TopicManager_retrieve(IceStorm_TopicManagerPrxPtr self,
                               Ice_String name,
                               IceStorm_TopicPrxPtr retval) {
    Ice_OutputStreamPtr os = &(self->stream);

    trace();
    
    Ptr_check(self);
    
    Ice_ObjectPrx_connect(self);
    Ice_OutputStream_init(os);
    
    /* request header */
    Ice_OutputStream_writeBlob(os, Ice_header, sizeof(Ice_header));
    
    /* request body */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 1);
    }
    if (self->mode != Ice_ReferenceModeTwoway) {
        Ice_OutputStream_writeInt(os, 0);
    }
    
    Ice_OutputStream_writeIdentity(os, self->identity);
    Ice_OutputStream_writeString(os, "");
    Ice_OutputStream_writeString(os, "retrieve");
    Ice_OutputStream_writeByte(os, Ice_Idempotent);
    Ice_OutputStream_writeByte(os, 0);
    
    /* encapsulated params */
    Ice_OutputStream_startWriteEncaps(os);
    Ice_OutputStream_writeIceString(os, name);
    Ice_OutputStream_endWriteEncaps(os);
    Ice_OutputStream_setMessageSize(os);
    
    Ice_ObjectPrx_send(self);
    
    /* parse reply message (only on twoway invocations) */
    if (self->mode == Ice_ReferenceModeTwoway) {
        Ice_InputStream is;
        char expected[10] = {'I', 'c', 'e', 'P', 1, 0, 1, 0, Ice_replyMsg, 0};
        Ice_Int size = 0;
        Ice_Int requestId = 0;
        Ice_Byte replyStatus = 0;
        Ice_Int encapsSize = 0;
        Ice_Byte skipByte = 0;
        
        uint16_t timeout = 5000;
        bool ready = false;
        while (!ready) {
            IcePlugin_Connection_dataReady(&(self->connection), &ready);
            IceUtil_delay(1);
            if (! --timeout) {
                c_throw("Ice_TimeoutException", __FILE__, __LINE__);
                return;
            }
        }
        
        /* reply header */
        Ice_InputStream_init(&is, self->connection.fd, self->connection.epinfo->type);
        if (memcmp(expected, is.next, 10) != 0) {
            c_throw("Ice_ProtocolException", __FILE__, __LINE__);
            return;
        }
        
        /* skip header from input stream */
        is.next += 10;
        Ice_InputStream_readInt(&is, &size);
            if (size > MAX_MESSAGE_SIZE) {
            c_throw("Ice_MemoryLimitException", __FILE__, __LINE__);
            return;
        }
        
        /* reply body */
        Ice_InputStream_readInt(&is, &requestId);
            Ice_InputStream_readByte(&is, &replyStatus);
            
        /* check replyStatus */
        switch (replyStatus) {
        case 0: {
            /* success */
            break;
        }
        case 1: {
            /* user exception */
            Ice_ObjectPrx_parseUserException(self, &is);
            return;
        }
        case 2: {
            c_throw("Ice_ObjectDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 3: {
            c_throw("Ice_FacetDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 4: {
            c_throw("Ice_OperationDoesNotExist", __FILE__, __LINE__);
            return;
        }
        case 5:
        case 6:
        case 7: {
            /* unkown exception (reason as string) */
            /* FIXME: unmarshall and print reason */
            c_throw("Ice_UnknownException", __FILE__, __LINE__);
            return;
        }}
        
        Ice_InputStream_readInt(&is, &encapsSize);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_InputStream_readByte(&is, &skipByte);
            Ice_ObjectPrx_readFromInputStream(retval, &is);
        }
    
    Ice_ObjectPrx_closeConnection(self);
}

